# -*- coding: utf-8 -*-
"default plugin for setting: set it in a simple dictionary"
# Copyright (C) 2013-2019 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from copy import deepcopy
from ...log import log


class Properties:
    __slots__ = ('_properties',
                 '_storage')

    def __init__(self, storage):
        # properties attribute: the name of a property enables this property
        # key is None for global properties
        self._properties = {}
        self._storage = storage

    # properties
    def setproperties(self, path, index, properties):
        log.debug('setproperties %s %s %s', path, index, properties)
        self._properties.setdefault(path, {})[index] = properties

    def getproperties(self, path, index, default_properties):
        if path not in self._properties:
            ret = frozenset(default_properties)
        else:
            ret = self._properties[path].get(index, frozenset(default_properties))
        log.debug('getproperties %s %s %s', path, index, ret)
        return ret

    def delproperties(self, path, index):
        log.debug('delproperties %s', path)
        if path in self._properties and index in self._properties[path]:
            del(self._properties[path][index])


    def exportation(self):
        """return all modified settings in a dictionary
        example: {'path1': set(['prop1', 'prop2'])}
        """
        return deepcopy(self._properties)

    def importation(self, properties):
        self._properties = properties


class Permissives:
    __slots__ = ('_permissives',
                 '_storage')

    def __init__(self, storage):
        # permissive properties
        self._permissives = {}
        self._storage = storage

    def setpermissives(self, path, index, permissives):
        log.debug('setpermissives %s %s', path, permissives)
        self._permissives.setdefault(path, {})[index] = permissives

    def getpermissives(self, path, index):
        if not path in self._permissives:
            ret = frozenset()
        else:
            ret = self._permissives[path].get(index, frozenset())
        log.debug('getpermissives %s %s', path, ret)
        return ret

    def delpermissive(self, path, index):
        log.debug('delpermissive %s', path)
        if path in self._permissives and index in self._permissives[path]:
            del(self._permissives[path][index])

    def exportation(self):
        """return all modified permissives in a dictionary
        example: {'path1': set(['perm1', 'perm2'])}
        """
        return deepcopy(self._permissives)

    def importation(self, permissives):
        self._permissives = permissives
